@@ -2,8 +2,7 @@
 Alternate implementation of RobotParser (alternative to standard library urllib.robotparser)
 
 Reference:
-Robots Exclusion Protocol (REP) draft-koster-rep-01
-https://tools.ietf.org/html/draft-koster-rep
+Robots Exclusion Protocol (REP) https://www.rfc-editor.org/rfc/rfc9309
 """
 
 from typing import Dict, Iterator, List, NamedTuple, Tuple, Type, TypeVar
@@ -24,17 +23,17 @@ RE_AGENT_TOKEN = re.compile(r"^[a-zA-Z_-]+$")
 # Note: the hash character ('#') needs to be escaped in VERBOSE mode, otherwise it would be
 # interpreted as a comment.
 RE_AGENT = re.compile(
-    r"^\s*user-agent\s*:\s*(?P<AGENT>\*|[a-zA-Z_-]+)\s*(?:\#.*)?$",
+    r"^\s*user-agent\s*:\s*(?P<AGENT>\*|[a-zA-Z_-]+)[^]s]*\s*(?:\#.*)?$",
     re.IGNORECASE | re.VERBOSE,
 )
 
 RE_SITEMAP = re.compile(
-    r"^\s*sitemap\s*:\s*(?P<SITEMAP>https?:\/\/[^\n\s]+)\s*$",
+    r"^\s*sitemap\s*:\s*(?P<SITEMAP>https?://[^\n\s]+)\s*$",
     re.IGNORECASE | re.VERBOSE,
 )
 
 # Product token in the user-agent line:
-RE_PRODUCT = re.compile(r"(?:^[a-zA-Z_-]+$|\*)")
+RE_PRODUCT = re.compile(r"^[a-zA-Z_-]+$|\*")
 
 # Rule allow
 RE_RULE_ALLOW = re.compile(
@@ -128,7 +127,7 @@ T = TypeVar("T", bound="RobotsParser")
 
 class RobotsParser:
     """Encapsulates functions and data to parse a robotstxt file and get
-    feedback on what a crawler is allowed to access to on a given web site.
+    feedback on what a crawler is allowed to access to on a given website.
     """
 
     AGENT_NOT_VALID = "User agent [%s] is not valid"
@@ -144,7 +143,7 @@ class RobotsParser:
         self.allow_all = False
         self._host = ""
         self._path = "/robots.txt"
-        self._time = 0  # Time the robots.txt is fetched
+        self._time = 0.0  # Time the robots.txt is fetched
 
     @property
     def errors(self):
@@ -164,7 +163,7 @@ class RobotsParser:
     @url.setter
     def url(self, url):
         """Set the url (example: 'https://www.example.com/robots.txt'), the path of the robots.txt
-        file (example: '/robots.txt') and the hostname (example 'www.example.com').
+        file, example: '/robots.txt', and the hostname, example 'www.example.com'.
 
         It discards the scheme ('http' or 'https') for compatibility with the Python standard
         library module 'urllib.robotparser'."""
@@ -210,20 +209,26 @@ class RobotsParser:
     def gen_uri(self, uri: str):
         """Instantiate a generator from a URI (either http:// or https:// or file:///"""
         try:
-            with urllib.request.urlopen(uri) as f:
+            req = urllib.request.Request(uri, headers={'User-Agent': 'RobotsPy'})
+            with urllib.request.urlopen(req) as f:
                 self._time = time.time()
                 for line in f:
-                    yield line.decode("utf-8")
+                    try:
+                        yield line.decode("utf-8")
+                    except UnicodeDecodeError as err:
+                        self.allow_all = True
+                        self._errors.append(("robots.txt must be UTF-8 encoded", f"{str(err)} for {uri}"))
+                        return
         except urllib.error.HTTPError as err:
             if err.code in (401, 403):
                 self.disallow_all = True
-                self._errors.append((str(err.code), str(err)))
+                self._errors.append((str(err.code), f"{str(err)} for {uri}"))
             elif 400 <= err.code < 500:
                 self.allow_all = True
-                self._warnings.append((str(err.code), str(err)))
+                self._warnings.append((str(err.code), f"{str(err)} for {uri}"))
             self._time = 0
         except urllib.error.URLError as err:
-            self._errors.append(("", str(err)))
+            self._errors.append(("", f"{str(err)} for {uri}"))
 
     @classmethod
     def from_uri(cls: Type[T], uri: str) -> T:
@@ -295,8 +300,7 @@ class RobotsParser:
                     current_rules = []
                 state = State.AGENT
                 current_agents.append(token.value.lower())
-            # Mypy issue https://github.com/python/mypy/issues/7316
-            elif token.type == TokenType.ALLOW or token.type == TokenType.DISALLOW:
+            elif token.type in (TokenType.ALLOW, TokenType.DISALLOW):
                 if state == State.BEGIN:
                     self._warnings.append(
                         (
@@ -333,7 +337,9 @@ class RobotsParser:
         rules for wild card agent ('*'), if no rule for '*', return empty list.
         """
 
-        rules = self.agents_rules.get(agent, self.agents_rules.get("*", []))
+        # Crawlers MUST use case-insensitive matching to find the group that matches the product token =>
+        # convert the product token (agent) to lower case.
+        rules = self.agents_rules.get(agent.lower(), self.agents_rules.get("*", []))
         return rules
 
     @staticmethod
@@ -361,7 +367,7 @@ class RobotsParser:
 
         # extract the path portion of the URL as-is (e.g. preserve a standalone ?)
         host_url = urllib.parse.urlunsplit((result.scheme, result.netloc, "", "", ""))
-        path = url[len(host_url) :] or "/"
+        path = url[len(host_url):] or "/"
         return result.netloc, RobotsParser.dedup_slash(path)
 
     @staticmethod
@@ -396,7 +402,7 @@ class RobotsParser:
         if host and self.host and host != self.host:
             return False
 
-        rules = self.find_rules(useragent.lower())
+        rules = self.find_rules(useragent)
 
         for rule in rules:
             # $ is a special character for robots and indicate the exact end of the pattern
